/* gcal-agenda-view.h
 *
 * Copyright (C) 2022 Purism SPC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "gcal-manager.h"
#include "gcal-view.h"

G_BEGIN_DECLS

#define GCAL_TYPE_AGENDA_VIEW (gcal_agenda_view_get_type ())
G_DECLARE_FINAL_TYPE (GcalAgendaView, gcal_agenda_view, GCAL, AGENDA_VIEW, GtkBox)

G_END_DECLS
